namespace com.progsbase.no.inductive.libraries.Unicode_v0_1_6{

using static System.Math;

public class ByteArray{
	public double [] bytes;
}
public class ByteArrayReference{
	public ByteArray bytes;
}
public class BooleanArrayReference{
	public bool [] booleanArray;
}
public class BooleanReference{
	public bool booleanValue;
}
public class CharacterReference{
	public char characterValue;
}
public class NumberArrayReference{
	public double [] numberArray;
}
public class NumberReference{
	public double numberValue;
}
public class StringArrayReference{
	public StringReference [] stringArray;
}
public class StringReference{
	public char [] stringx;
}
public class Unicode{
	public static ByteArray UnicodeToUTF8(double [] unicode){
		ByteArray utf8;
		double length, i, pos, p1, p2, p3, p4, code;

		length = 0d;
		for(i = 0d; i < unicode.Length; i = i + 1d){
			code = unicode[(int)(i)];

			length = length + 1d;
			if(code >= 128d){
				length = length + 1d;
			}
			if(code >= 2048d){
				length = length + 1d;
			}
			if(code >= 65536d){
				length = length + 1d;
			}
		}

		utf8 = CreateByteArray(length);

		pos = 0d;
		for(i = 0d; i < unicode.Length; i = i + 1d){
			code = unicode[(int)(i)];

			if(code < 128d){
				SetByte(utf8, pos, code);
				pos = pos + 1d;
			}else if(code < 2048d){
				p1 = code%64d;
				p2 = Floor(code/64d)%64d;

				SetByte(utf8, pos, 128d + 64d + p2);
				pos = pos + 1d;
				SetByte(utf8, pos, 128d + p1);
				pos = pos + 1d;
			}else if(code < 65536d){
				p1 = code%64d;
				p2 = Floor(code/64d)%64d;
				p3 = Floor(code/4096d);

				SetByte(utf8, pos, 128d + 64d + 32d + p3);
				pos = pos + 1d;
				SetByte(utf8, pos, 128d + p2);
				pos = pos + 1d;
				SetByte(utf8, pos, 128d + p1);
				pos = pos + 1d;
			}else{
				p1 = code%64d;
				p2 = Floor(code/64d)%64d;
				p3 = Floor(code/4096d)%64d;
				p4 = Floor(code/262144d);

				SetByte(utf8, pos, 128d + 64d + 32d + 16d + p4);
				pos = pos + 1d;
				SetByte(utf8, pos, 128d + p3);
				pos = pos + 1d;
				SetByte(utf8, pos, 128d + p2);
				pos = pos + 1d;
				SetByte(utf8, pos, 128d + p1);
				pos = pos + 1d;
			}
		}

		return utf8;
	}


	public static double [] UTF8ToUnicode(ByteArray utf8){
		double [] utf;
		double length, i, pos, p1, p2, p3, p4, code;

		length = 0d;
		for(i = 0d; i < ByteArrayLength(utf8); ){
			code = Floor(GetByte(utf8, i)/8d)*8d;

			if(code < 128d){
				length = length + 1d;
				i = i + 1d;
			}else if(code == 128d + 64d){
				length = length + 1d;
				i = i + 2d;
			}else if(code == 128d + 64d + 32d){
				length = length + 1d;
				i = i + 3d;
			}else if(code == 128d + 64d + 32d + 16d){
				length = length + 1d;
				i = i + 4d;
			}
		}

		utf = new double [(int)(length)];

		pos = 0d;
		for(i = 0d; i < ByteArrayLength(utf8); ){
			code = Floor(GetByte(utf8, i)/8d)*8d;

			if(code < 128d){
				utf[(int)(pos)] = GetByte(utf8, i);
				i = i + 1d;
				pos = pos + 1d;
			}else if(code == 128d + 64d){
				p1 = GetByte(utf8, i + 1d) - 128d;
				p2 = GetByte(utf8, i) - 128d - 64d;

				utf[(int)(pos)] = p2*64d + p1;

				i = i + 2d;
				pos = pos + 1d;
			}else if(code == 128d + 64d + 32d){
				p1 = GetByte(utf8, i + 2d) - 128d;
				p2 = GetByte(utf8, i + 1d) - 128d;
				p3 = GetByte(utf8, i) - 128d - 64d - 32d;

				utf[(int)(pos)] = p3*4096d + p2*64d + p1;

				i = i + 3d;
				pos = pos + 1d;
			}else if(code == 128d + 64d + 32d + 16d){
				p1 = GetByte(utf8, i + 3d) - 128d;
				p2 = GetByte(utf8, i + 2d) - 128d;
				p3 = GetByte(utf8, i + 1d) - 128d;
				p4 = GetByte(utf8, i) - 128d - 64d - 32d - 16d;

				utf[(int)(pos)] = p4*262144d + p3*4096d + p2*64d + p1;

				i = i + 4d;
				pos = pos + 1d;
			}
		}

		return utf;
	}


	public static double [] UTF16ToUnicode(char [] utf16){
		double [] unicode;
		double length, i, pos, hs, ls, code;
		double surrogateLowerBound, surrogateUpperBound, highSurrogateBase, lowSurrogateBase, highMultiplier, surrogateBase;

		surrogateLowerBound = 55296d;
		surrogateUpperBound = 57344d;
		highSurrogateBase = 55296d;
		lowSurrogateBase = 56320d;
		highMultiplier = 1024d;
		surrogateBase = 65536d;

		length = 0d;
		for(i = 0d; i < utf16.Length; ){
			code = utf16[(int)(i)];

			if(code < surrogateLowerBound || code >= surrogateUpperBound){
				length = length + 1d;
				i = i + 1d;
			}else{
				length = length + 1d;
				i = i + 2d;
			}
		}

		unicode = new double [(int)(length)];

		pos = 0d;
		for(i = 0d; i < utf16.Length; ){
			code = utf16[(int)(i)];

			if(code < surrogateLowerBound || code >= surrogateUpperBound){
				unicode[(int)(pos)] = utf16[(int)(i)];
				i = i + 1d;
				pos = pos + 1d;
			}else{
				hs = utf16[(int)(i)];
				ls = utf16[(int)(i + 1d)];

				unicode[(int)(pos)] = (hs - highSurrogateBase)*highMultiplier + (ls - lowSurrogateBase) + surrogateBase;

				i = i + 2d;
				pos = pos + 1d;
			}
		}

		return unicode;
	}


	public static char [] UnicodeToUTF16(double [] unicode){
		char [] utf16;
		double length, code, i, pos, hs, ls;
		double highSurrogateBase, lowSurrogateBase, highMultiplier, surrogateBase;

		highSurrogateBase = 55296d;
		lowSurrogateBase = 56320d;
		highMultiplier = 1024d;
		surrogateBase = 65536d;

		length = 0d;
		for(i = 0d; i < unicode.Length; i = i + 1d){
			if(unicode[(int)(i)] < surrogateBase){
				length = length + 1d;
			}else{
				length = length + 2d;
			}
		}

		utf16 = new char [(int)(length)];

		pos = 0d;
		for(i = 0d; i < unicode.Length; i = i + 1d){
			if(unicode[(int)(i)] < surrogateBase){
				utf16[(int)(pos)] = (char)(unicode[(int)(i)]);
				pos = pos + 1d;
			}else{
				code = unicode[(int)(i)] - surrogateBase;
				hs = Floor(code/highMultiplier) + highSurrogateBase;
				ls = code%highMultiplier + lowSurrogateBase;

				utf16[(int)(pos)] = (char)(hs);
				pos = pos + 1d;
				utf16[(int)(pos)] = (char)(ls);
				pos = pos + 1d;
			}
		}

		return utf16;
	}


	public static char [] UTF8ToUTF16(ByteArray utf8){
		double [] unicode;
		char [] utf16;

		unicode = UTF8ToUnicode(utf8);
		utf16 = UnicodeToUTF16(unicode);

		delete(unicode);

		return utf16;
	}


	public static ByteArray UTF16ToUTF8(char [] utf16){
		double [] unicode;
		ByteArray utf8;

		unicode = UTF16ToUnicode(utf16);
		utf8 = UnicodeToUTF8(unicode);

		delete(unicode);

		return utf8;
	}


	public static bool IsValidUnicode(double [] unicode){
		double lower, upper, i, code;
		bool valid;

		lower = 0d;
		upper = 1114112d;

		valid = true;
		for(i = 0d; i < unicode.Length && valid; i = i + 1d){
			code = unicode[(int)(i)];

			if(code == Floor(code)){
				if(code >= lower && code < upper){
				}else{
					valid = false;
				}
			}else{
				valid = false;
			}
		}

		return valid;
	}


	public static bool IsValidUtf16(char [] utf16){
		double i, code;
		bool valid;
		double surrogateLowerBound, surrogateUpperBound;

		surrogateLowerBound = 55296d;
		surrogateUpperBound = 57344d;

		valid = true;
		for(i = 0d; i < utf16.Length && valid; i = i + 1d){
			code = utf16[(int)(i)];

			if(code == Floor(code)){
				if(code < surrogateLowerBound || code >= surrogateUpperBound){
				}else if(i + 1d < utf16.Length){
					i = i + 1d;
					code = utf16[(int)(i)];

					if(code < surrogateLowerBound || code >= surrogateUpperBound){
						valid = false;
					}
				}else{
					valid = false;
				}
			}else{
				valid = false;
			}
		}

		return valid;
	}


	public static bool IsValidUtf8(ByteArray utf8){
		double i, code1, code2, code3, code4;
		bool valid;

		valid = true;
		for(i = 0d; i < ByteArrayLength(utf8) && valid; i = i + 1d){
			code1 = GetByte(utf8, i);

			if(code1 == Floor(code1)){
				if(code1 < 128d){
				}else if(code1 < 128d + 64d){
					valid = false;
				}else if(code1 < 128d + 64d + 32d){
					if(i + 1d < ByteArrayLength(utf8)){
						i = i + 1d;
						code2 = GetByte(utf8, i);

						if(code2 >= 128d && code2 < 128d + 64d){
						}else{
							valid = false;
						}
					}else{
						valid = false;
					}
				}else if(code1 < 128d + 64d + 32d + 16d){
					if(i + 2d < ByteArrayLength(utf8)){
						i = i + 1d;
						code2 = GetByte(utf8, i);
						i = i + 1d;
						code3 = GetByte(utf8, i);

						if(code2 >= 128d && code2 < 128d + 64d && code3 >= 128d && code3 < 128d + 64d){
						}else{
							valid = false;
						}
					}else{
						valid = false;
					}
				}else if(i + 3d < ByteArrayLength(utf8)){
					i = i + 1d;
					code2 = GetByte(utf8, i);
					i = i + 1d;
					code3 = GetByte(utf8, i);
					i = i + 1d;
					code4 = GetByte(utf8, i);

					if(code2 >= 128d && code2 < 128d + 64d && code3 >= 128d && code3 < 128d + 64d && code4 >= 128d && code4 < 128d + 64d){
					}else{
						valid = false;
					}
				}else{
					valid = false;
				}
			}else{
				valid = false;
			}
		}

		return valid;
	}


	public static ByteArray ReadXbytes(ByteArray data, NumberReference position, double length){
		ByteArray r;
		double i;

		r = CreateByteArray(length);

		for(i = 0d; i < length; i = i + 1d){
			r.bytes[(int)(i)] = ReadByte(data, position);
		}

		return r;
	}


	public static double Read4bytesBE(ByteArray data, NumberReference position){
		double r;

		r = 0d;
		r = r + Pow(2d, 24d)*ReadByte(data, position);
		r = r + Pow(2d, 16d)*ReadByte(data, position);
		r = r + Pow(2d, 8d)*ReadByte(data, position);
		r = r + ReadByte(data, position);

		return r;
	}


	public static double Read2bytesBE(ByteArray data, NumberReference position){
		double r;

		r = 0d;
		r = r + Pow(2d, 8d)*ReadByte(data, position);
		r = r + ReadByte(data, position);

		return r;
	}


	public static double ReadByte(ByteArray data, NumberReference position){
		double next;

		next = data.bytes[(int)(position.numberValue)];
		position.numberValue = position.numberValue + 1d;

		return next;
	}


	public static double Read4bytesLE(ByteArray data, NumberReference position){
		double r;

		r = 0d;
		r = r + ReadByte(data, position);
		r = r + Pow(2d, 8d)*ReadByte(data, position);
		r = r + Pow(2d, 16d)*ReadByte(data, position);
		r = r + Pow(2d, 24d)*ReadByte(data, position);

		return r;
	}


	public static void WriteByte(ByteArray data, double b, NumberReference position){
		data.bytes[(int)(position.numberValue)] = b;
		position.numberValue = position.numberValue + 1d;
	}


	public static void Write2BytesLE(ByteArray data, double b, NumberReference position){
		data.bytes[(int)(position.numberValue)] = BytesRound(b%Pow(2d, 8d));
		position.numberValue = position.numberValue + 1d;
		data.bytes[(int)(position.numberValue)] = Floor(b/Pow(2d, 8d))%Pow(2d, 8d);
		position.numberValue = position.numberValue + 1d;
	}


	public static void Write4BytesLE(ByteArray data, double b, NumberReference position){
		data.bytes[(int)(position.numberValue)] = BytesRound(b%Pow(2d, 8d));
		position.numberValue = position.numberValue + 1d;
		data.bytes[(int)(position.numberValue)] = Floor(b/Pow(2d, 8d))%Pow(2d, 8d);
		position.numberValue = position.numberValue + 1d;
		data.bytes[(int)(position.numberValue)] = Floor(b/Pow(2d, 16d))%Pow(2d, 8d);
		position.numberValue = position.numberValue + 1d;
		data.bytes[(int)(position.numberValue)] = Floor(b/Pow(2d, 24d))%Pow(2d, 8d);
		position.numberValue = position.numberValue + 1d;
	}


	public static void Write2BytesBE(ByteArray data, double b, NumberReference position){
		data.bytes[(int)(position.numberValue)] = Floor(b/Pow(2d, 8d))%Pow(2d, 8d);
		position.numberValue = position.numberValue + 1d;
		data.bytes[(int)(position.numberValue)] = BytesRound(b%Pow(2d, 8d));
		position.numberValue = position.numberValue + 1d;
	}


	public static void Write4BytesBE(ByteArray data, double b, NumberReference position){
		data.bytes[(int)(position.numberValue)] = Floor(b/Pow(2d, 24d))%Pow(2d, 8d);
		position.numberValue = position.numberValue + 1d;
		data.bytes[(int)(position.numberValue)] = Floor(b/Pow(2d, 16d))%Pow(2d, 8d);
		position.numberValue = position.numberValue + 1d;
		data.bytes[(int)(position.numberValue)] = Floor(b/Pow(2d, 8d))%Pow(2d, 8d);
		position.numberValue = position.numberValue + 1d;
		data.bytes[(int)(position.numberValue)] = BytesRound(b%Pow(2d, 8d));
		position.numberValue = position.numberValue + 1d;
	}


	public static void WriteStringBytes(ByteArray data, char [] cs, NumberReference position){
		double i, v;

		for(i = 0d; i < cs.Length; i = i + 1d){
			v = cs[(int)(i)];
			WriteByte(data, v, position);
		}
	}


	public static double BytesRound(double x){
		return Floor(x + 0.5);
	}


	public static double [] ByteArrayToNumberArray(ByteArray src){
		double [] dst;
		double i, len;

		len = src.bytes.Length;
		dst = new double [(int)(len)];
		for(i = 0d; i < len; i = i + 1d){
			dst[(int)(i)] = src.bytes[(int)(i)];
		}

		return dst;
	}


	public static ByteArray NumberArrayToByteArray(double [] src){
		ByteArray dst;
		double i, len;

		len = src.Length;
		dst = new ByteArray();
		dst.bytes = new double [(int)(len)];
		for(i = 0d; i < len; i = i + 1d){
			dst.bytes[(int)(i)] = src[(int)(i)];
		}

		return dst;
	}


	public static bool ByteArraysEqual(ByteArray a, ByteArray b){
		bool equal;
		double i;

		equal = true;
		if(a.bytes.Length == b.bytes.Length){
			for(i = 0d; i < a.bytes.Length && equal; i = i + 1d){
				if(a.bytes[(int)(i)] != b.bytes[(int)(i)]){
					equal = false;
				}
			}
		}else{
			equal = false;
		}

		return equal;
	}


	public static ByteArray CopyByteArray(ByteArray a){
		double i;
		ByteArray n;

		n = new ByteArray();
		n.bytes = new double [(int)(a.bytes.Length)];

		for(i = 0d; i < a.bytes.Length; i = i + 1d){
			n.bytes[(int)(i)] = a.bytes[(int)(i)];
		}

		return n;
	}


	public static double ByteArrayLength(ByteArray response){
		return response.bytes.Length;
	}


	public static ByteArray CreateAndFillByteArray(double length, double value){
		ByteArray bytes;
		double i;

		bytes = new ByteArray();
		bytes.bytes = new double [(int)(length)];

		for(i = 0d; i < length; i = i + 1d){
			bytes.bytes[(int)(i)] = value;
		}

		return bytes;
	}


	public static ByteArray CreateByteArray(double length){
		ByteArray bytes;

		bytes = new ByteArray();
		bytes.bytes = new double [(int)(length)];

		return bytes;
	}


	public static void SetByte(ByteArray array, double index, double value){
		array.bytes[(int)(index)] = value;
	}


	public static double GetByte(ByteArray array, double index){
		return array.bytes[(int)(index)];
	}


	public static void AssertByteArraysEqual(ByteArray a, ByteArray b, NumberReference failures){
		double i;

		if(a.bytes.Length == b.bytes.Length){
			for(i = 0d; i < a.bytes.Length; i = i + 1d){
				if(GetByte(a, i) != GetByte(b, i)){
					failures.numberValue = failures.numberValue + 1d;
				}
			}
		}else{
			failures.numberValue = failures.numberValue + 1d;
		}
	}


	public static void FreeByteArray(ByteArray byteArray){
		delete(byteArray.bytes);
		delete(byteArray);
	}


	public static bool CopyByteArrayRange(ByteArray a, double from, double to, ByteArray b){
		double i, length;
		double [] n;
		bool success;

		if(from >= 0d && from <= a.bytes.Length && to >= 0d && to <= a.bytes.Length && from <= to){
			length = to - from;
			n = new double [(int)(length)];

			for(i = 0d; i < length; i = i + 1d){
				n[(int)(i)] = a.bytes[(int)(from + i)];
			}

			b.bytes = n;
			success = true;
		}else{
			success = false;
		}

		return success;
	}


	public static char [] BytesToTextBase16(ByteArray bytes){
		StringReference [] comb;
		double i, v;
		char [] r;

		r = new char [(int)(bytes.bytes.Length*2d)];
		comb = GenerateBase16ByteCombinations();

		for(i = 0d; i < bytes.bytes.Length; i = i + 1d){
			v = bytes.bytes[(int)(i)];
			r[(int)(2d*i + 0d)] = comb[(int)(v)].stringx[0];
			r[(int)(2d*i + 1d)] = comb[(int)(v)].stringx[1];
		}

		FreeBase64ByteCombinations(comb);

		return r;
	}


	public static bool TextToBytesBase16WithCheck(char [] stringx, ByteArrayReference bytes, StringReference message){
		StringReference [] comb;
		double i, n, c, d;
		double [] reverseBase16;
		char [] codes;
		bool success;

		codes = "0123456789ABCDEF".ToCharArray();

		reverseBase16 = new double [128];
		for(i = 0d; i < codes.Length; i = i + 1d){
			n = codes[(int)(i)];
			reverseBase16[(int)(n)] = i;
		}

		if(stringx.Length%2d == 0d){
			bytes.bytes = CreateByteArray(stringx.Length/2d);
			comb = GenerateBase16ByteCombinations();

			success = true;
			for(i = 0d; i < stringx.Length/2d && success; i = i + 1d){
				c = stringx[(int)(2d*i + 0d)];
				d = stringx[(int)(2d*i + 1d)];

				if(c == 'a'){
					c = 'A';
				}
				if(c == 'b'){
					c = 'B';
				}
				if(c == 'c'){
					c = 'C';
				}
				if(c == 'd'){
					c = 'D';
				}
				if(c == 'e'){
					c = 'E';
				}
				if(c == 'f'){
					c = 'F';
				}

				if(d == 'a'){
					d = 'A';
				}
				if(d == 'b'){
					d = 'B';
				}
				if(d == 'c'){
					d = 'C';
				}
				if(d == 'd'){
					d = 'D';
				}
				if(d == 'e'){
					d = 'E';
				}
				if(d == 'f'){
					d = 'F';
				}

				success = true;
				success = success && (c == '0' || c == '1' || c == '2' || c == '3' || c == '4' || c == '5' || c == '6' || c == '7' || c == '8' || c == '9' || c == 'A' || c == 'B' || c == 'C' || c == 'D' || c == 'E' || c == 'F');
				success = success && (d == '0' || d == '1' || d == '2' || d == '3' || d == '4' || d == '5' || d == '6' || d == '7' || d == '8' || d == '9' || d == 'A' || d == 'B' || d == 'C' || d == 'D' || d == 'E' || d == 'F');

				if(success){
					n = reverseBase16[(int)(c)]*16d + reverseBase16[(int)(d)];

					bytes.bytes.bytes[(int)(i)] = n;
				}else{
					success = false;
					message.stringx = "Hex-string can only contain 0-9, A-F or a-f.".ToCharArray();
				}
			}

			FreeBase64ByteCombinations(comb);
		}else{
			bytes.bytes = CreateByteArray(0d);
			success = false;
			message.stringx = "Hex-string must be an even number of characters.".ToCharArray();
		}

		delete(reverseBase16);

		return success;
	}


	public static ByteArray TextToBytesBase16(char [] stringx){
		StringReference [] comb;
		double i, n, c, d;
		ByteArray r;
		double [] reverseBase16;
		char [] codes;

		codes = "0123456789ABCDEF".ToCharArray();

		reverseBase16 = new double [128];
		for(i = 0d; i < codes.Length; i = i + 1d){
			n = codes[(int)(i)];
			reverseBase16[(int)(n)] = i;
		}

		if(stringx.Length%2d == 0d){
			r = CreateByteArray(stringx.Length/2d);
			comb = GenerateBase16ByteCombinations();

			for(i = 0d; i < stringx.Length/2d; i = i + 1d){
				c = stringx[(int)(2d*i + 0d)];
				d = stringx[(int)(2d*i + 1d)];

				n = reverseBase16[(int)(c)]*16d + reverseBase16[(int)(d)];

				r.bytes[(int)(i)] = n;
			}

			FreeBase64ByteCombinations(comb);
		}else{
			r = CreateByteArray(0d);
		}

		delete(reverseBase16);

		return r;
	}


	public static void FreeBase64ByteCombinations(StringReference [] comb){
		double i;
		for(i = 0d; i < 256d; i = i + 1d){
			FreeStringReference(comb[(int)(i)]);
		}
		delete(comb);
	}


	public static StringReference [] GenerateBase16ByteCombinations(){
		StringReference [] comb;
		char [] codes;
		double i, j;
		comb = new StringReference [256];

		codes = "0123456789ABCDEF".ToCharArray();

		for(i = 0d; i < 256d; i = i + 1d){
			comb[(int)(i)] = new StringReference();
			comb[(int)(i)].stringx = new char [2];
		}

		for(j = 0d; j < 16d; j = j + 1d){
			for(i = 0d; i < 16d; i = i + 1d){
				comb[(int)(j*16d + i)].stringx[1] = codes[(int)(i)];
				comb[(int)(j*16d + i)].stringx[0] = codes[(int)(j)];
			}
		}

		return comb;
	}


	public static BooleanReference CreateBooleanReference(bool value){
		BooleanReference refx;

		refx = new BooleanReference();
		refx.booleanValue = value;

		return refx;
	}


	public static BooleanArrayReference CreateBooleanArrayReference(bool [] value){
		BooleanArrayReference refx;

		refx = new BooleanArrayReference();
		refx.booleanArray = value;

		return refx;
	}


	public static BooleanArrayReference CreateBooleanArrayReferenceLengthValue(double length, bool value){
		BooleanArrayReference refx;
		double i;

		refx = new BooleanArrayReference();
		refx.booleanArray = new bool [(int)(length)];

		for(i = 0d; i < length; i = i + 1d){
			refx.booleanArray[(int)(i)] = value;
		}

		return refx;
	}


	public static void FreeBooleanArrayReference(BooleanArrayReference booleanArrayReference){
		delete(booleanArrayReference.booleanArray);
		delete(booleanArrayReference);
	}


	public static CharacterReference CreateCharacterReference(char value){
		CharacterReference refx;

		refx = new CharacterReference();
		refx.characterValue = value;

		return refx;
	}


	public static NumberReference CreateNumberReference(double value){
		NumberReference refx;

		refx = new NumberReference();
		refx.numberValue = value;

		return refx;
	}


	public static NumberArrayReference CreateNumberArrayReference(double [] value){
		NumberArrayReference refx;

		refx = new NumberArrayReference();
		refx.numberArray = value;

		return refx;
	}


	public static NumberArrayReference CreateNumberArrayReferenceLengthValue(double length, double value){
		NumberArrayReference refx;
		double i;

		refx = new NumberArrayReference();
		refx.numberArray = new double [(int)(length)];

		for(i = 0d; i < length; i = i + 1d){
			refx.numberArray[(int)(i)] = value;
		}

		return refx;
	}


	public static void FreeNumberArrayReference(NumberArrayReference numberArrayReference){
		delete(numberArrayReference.numberArray);
		delete(numberArrayReference);
	}


	public static StringReference CreateStringReference(char [] value){
		StringReference refx;

		refx = new StringReference();
		refx.stringx = value;

		return refx;
	}


	public static StringReference CreateStringReferenceLengthValue(double length, char value){
		StringReference refx;
		double i;

		refx = new StringReference();
		refx.stringx = new char [(int)(length)];

		for(i = 0d; i < length; i = i + 1d){
			refx.stringx[(int)(i)] = value;
		}

		return refx;
	}


	public static void FreeStringReference(StringReference stringReference){
		delete(stringReference.stringx);
		delete(stringReference);
	}


	public static StringArrayReference CreateStringArrayReference(StringReference [] strings){
		StringArrayReference refx;

		refx = new StringArrayReference();
		refx.stringArray = strings;

		return refx;
	}


	public static StringArrayReference CreateStringArrayReferenceLengthValue(double length, char [] value){
		StringArrayReference refx;
		double i;

		refx = new StringArrayReference();
		refx.stringArray = new StringReference [(int)(length)];

		for(i = 0d; i < length; i = i + 1d){
			refx.stringArray[(int)(i)] = CreateStringReference(value);
		}

		return refx;
	}


	public static void FreeStringArrayReference(StringArrayReference stringArrayReference){
		double i;

		for(i = 0d; i < stringArrayReference.stringArray.Length; i = i + 1d){
			delete(stringArrayReference.stringArray[(int)(i)]);
		}
		delete(stringArrayReference.stringArray);
		delete(stringArrayReference);
	}


	public static void AssertFalse(bool b, NumberReference failures){
		if(b){
			failures.numberValue = failures.numberValue + 1d;
		}
	}


	public static void AssertTrue(bool b, NumberReference failures){
		if(!b){
			failures.numberValue = failures.numberValue + 1d;
		}
	}


	public static void AssertEquals(double a, double b, NumberReference failures){
		if(a != b){
			failures.numberValue = failures.numberValue + 1d;
		}
	}


	public static void AssertBooleansEqual(bool a, bool b, NumberReference failures){
		if(a != b){
			failures.numberValue = failures.numberValue + 1d;
		}
	}


	public static void AssertCharactersEqual(char a, char b, NumberReference failures){
		if(a != b){
			failures.numberValue = failures.numberValue + 1d;
		}
	}


	public static void AssertStringEquals(char [] a, char [] b, NumberReference failures){
		if(!StringsEqual(a, b)){
			failures.numberValue = failures.numberValue + 1d;
		}
	}


	public static void AssertNumberArraysEqual(double [] a, double [] b, NumberReference failures){
		double i;

		if(a.Length == b.Length){
			for(i = 0d; i < a.Length; i = i + 1d){
				AssertEquals(a[(int)(i)], b[(int)(i)], failures);
			}
		}else{
			failures.numberValue = failures.numberValue + 1d;
		}
	}


	public static void AssertBooleanArraysEqual(bool [] a, bool [] b, NumberReference failures){
		double i;

		if(a.Length == b.Length){
			for(i = 0d; i < a.Length; i = i + 1d){
				AssertBooleansEqual(a[(int)(i)], b[(int)(i)], failures);
			}
		}else{
			failures.numberValue = failures.numberValue + 1d;
		}
	}


	public static void AssertStringArraysEqual(StringReference [] a, StringReference [] b, NumberReference failures){
		double i;

		if(a.Length == b.Length){
			for(i = 0d; i < a.Length; i = i + 1d){
				AssertStringEquals(a[(int)(i)].stringx, b[(int)(i)].stringx, failures);
			}
		}else{
			failures.numberValue = failures.numberValue + 1d;
		}
	}


	public static double [] StringToNumberArray(char [] stringx){
		double i;
		double [] array;

		array = new double [(int)(stringx.Length)];

		for(i = 0d; i < stringx.Length; i = i + 1d){
			array[(int)(i)] = stringx[(int)(i)];
		}
		return array;
	}


	public static char [] NumberArrayToString(double [] array){
		double i;
		char [] stringx;

		stringx = new char [(int)(array.Length)];

		for(i = 0d; i < array.Length; i = i + 1d){
			stringx[(int)(i)] = (char)(array[(int)(i)]);
		}
		return stringx;
	}


	public static bool NumberArraysEqual(double [] a, double [] b){
		bool equal;
		double i;

		equal = true;
		if(a.Length == b.Length){
			for(i = 0d; i < a.Length && equal; i = i + 1d){
				if(a[(int)(i)] != b[(int)(i)]){
					equal = false;
				}
			}
		}else{
			equal = false;
		}

		return equal;
	}


	public static bool BooleanArraysEqual(bool [] a, bool [] b){
		bool equal;
		double i;

		equal = true;
		if(a.Length == b.Length){
			for(i = 0d; i < a.Length && equal; i = i + 1d){
				if(a[(int)(i)] != b[(int)(i)]){
					equal = false;
				}
			}
		}else{
			equal = false;
		}

		return equal;
	}


	public static bool StringsEqual(char [] a, char [] b){
		bool equal;
		double i;

		equal = true;
		if(a.Length == b.Length){
			for(i = 0d; i < a.Length && equal; i = i + 1d){
				if(a[(int)(i)] != b[(int)(i)]){
					equal = false;
				}
			}
		}else{
			equal = false;
		}

		return equal;
	}


	public static void FillNumberArray(double [] a, double value){
		double i;

		for(i = 0d; i < a.Length; i = i + 1d){
			a[(int)(i)] = value;
		}
	}


	public static void FillString(char [] a, char value){
		double i;

		for(i = 0d; i < a.Length; i = i + 1d){
			a[(int)(i)] = value;
		}
	}


	public static void FillBooleanArray(bool [] a, bool value){
		double i;

		for(i = 0d; i < a.Length; i = i + 1d){
			a[(int)(i)] = value;
		}
	}


	public static bool FillNumberArrayRange(double [] a, double value, double from, double to){
		double i, length;
		bool success;

		if(from >= 0d && from <= a.Length && to >= 0d && to <= a.Length && from <= to){
			length = to - from;
			for(i = 0d; i < length; i = i + 1d){
				a[(int)(from + i)] = value;
			}

			success = true;
		}else{
			success = false;
		}

		return success;
	}


	public static bool FillBooleanArrayRange(bool [] a, bool value, double from, double to){
		double i, length;
		bool success;

		if(from >= 0d && from <= a.Length && to >= 0d && to <= a.Length && from <= to){
			length = to - from;
			for(i = 0d; i < length; i = i + 1d){
				a[(int)(from + i)] = value;
			}

			success = true;
		}else{
			success = false;
		}

		return success;
	}


	public static bool FillStringRange(char [] a, char value, double from, double to){
		double i, length;
		bool success;

		if(from >= 0d && from <= a.Length && to >= 0d && to <= a.Length && from <= to){
			length = to - from;
			for(i = 0d; i < length; i = i + 1d){
				a[(int)(from + i)] = value;
			}

			success = true;
		}else{
			success = false;
		}

		return success;
	}


	public static double [] CopyNumberArray(double [] a){
		double i;
		double [] n;

		n = new double [(int)(a.Length)];

		for(i = 0d; i < a.Length; i = i + 1d){
			n[(int)(i)] = a[(int)(i)];
		}

		return n;
	}


	public static bool [] CopyBooleanArray(bool [] a){
		double i;
		bool [] n;

		n = new bool [(int)(a.Length)];

		for(i = 0d; i < a.Length; i = i + 1d){
			n[(int)(i)] = a[(int)(i)];
		}

		return n;
	}


	public static char [] CopyString(char [] a){
		double i;
		char [] n;

		n = new char [(int)(a.Length)];

		for(i = 0d; i < a.Length; i = i + 1d){
			n[(int)(i)] = a[(int)(i)];
		}

		return n;
	}


	public static bool CopyNumberArrayRange(double [] a, double from, double to, NumberArrayReference copyReference){
		double i, length;
		double [] n;
		bool success;

		if(from >= 0d && from <= a.Length && to >= 0d && to <= a.Length && from <= to){
			length = to - from;
			n = new double [(int)(length)];

			for(i = 0d; i < length; i = i + 1d){
				n[(int)(i)] = a[(int)(from + i)];
			}

			copyReference.numberArray = n;
			success = true;
		}else{
			success = false;
		}

		return success;
	}


	public static bool CopyBooleanArrayRange(bool [] a, double from, double to, BooleanArrayReference copyReference){
		double i, length;
		bool [] n;
		bool success;

		if(from >= 0d && from <= a.Length && to >= 0d && to <= a.Length && from <= to){
			length = to - from;
			n = new bool [(int)(length)];

			for(i = 0d; i < length; i = i + 1d){
				n[(int)(i)] = a[(int)(from + i)];
			}

			copyReference.booleanArray = n;
			success = true;
		}else{
			success = false;
		}

		return success;
	}


	public static bool CopyStringRange(char [] a, double from, double to, StringReference copyReference){
		double i, length;
		char [] n;
		bool success;

		if(from >= 0d && from <= a.Length && to >= 0d && to <= a.Length && from <= to){
			length = to - from;
			n = new char [(int)(length)];

			for(i = 0d; i < length; i = i + 1d){
				n[(int)(i)] = a[(int)(from + i)];
			}

			copyReference.stringx = n;
			success = true;
		}else{
			success = false;
		}

		return success;
	}


	public static bool IsLastElement(double length, double index){
		return index + 1d == length;
	}


	public static double [] CreateNumberArray(double length, double value){
		double [] array;

		array = new double [(int)(length)];
		FillNumberArray(array, value);

		return array;
	}


	public static bool [] CreateBooleanArray(double length, bool value){
		bool [] array;

		array = new bool [(int)(length)];
		FillBooleanArray(array, value);

		return array;
	}


	public static char [] CreateString(double length, char value){
		char [] array;

		array = new char [(int)(length)];
		FillString(array, value);

		return array;
	}


	public static void SwapElementsOfNumberArray(double [] A, double ai, double bi){
		double tmp;

		tmp = A[(int)(ai)];
		A[(int)(ai)] = A[(int)(bi)];
		A[(int)(bi)] = tmp;
	}


	public static void SwapElementsOfStringArray(StringArrayReference A, double ai, double bi){
		StringReference tmp;

		tmp = A.stringArray[(int)(ai)];
		A.stringArray[(int)(ai)] = A.stringArray[(int)(bi)];
		A.stringArray[(int)(bi)] = tmp;
	}


	public static void ReverseNumberArray(double [] array){
		double i;

		for(i = 0d; i < array.Length/2d; i = i + 1d){
			SwapElementsOfNumberArray(array, i, array.Length - i - 1d);
		}
	}


	public static void delete(System.Object objectx){
		// C# has garbage collection.
	}
}

}